import os
from typing import Any
from typing import cast
from typing import Dict
from typing import Iterable

from ...exceptions import InvalidFile
from ..plugins.util import get_plugins_from_file


def upgrade(baseline: Dict[str, Any]) -> None:
    for function in [
        _migrate_filters,
        _rename_high_entropy_string_arguments,
        _migrate_custom_plugins,
    ]:
        function(baseline)


def _migrate_filters(baseline: Dict[str, Any]) -> None:
    """
    In v1.0.0, we introduced the idea of `filters`. This consolidated a variety of different
    false positive filters into a configurable layout. To reduce upgrade friction, this will
    contain the default filters used before this version upgrade.
    """
    baseline['filters_used'] = [
        {
            'path': 'detect_secrets.filters.allowlist.is_line_allowlisted',
        },
        {
            'path': 'detect_secrets.filters.heuristic.is_sequential_string',
        },
        {
            'path': 'detect_secrets.filters.heuristic.is_potential_uuid',
        },
        {
            'path': 'detect_secrets.filters.heuristic.is_likely_id_string',
        },
        {
            'path': 'detect_secrets.filters.heuristic.is_templated_secret',
        },
        {
            'path': 'detect_secrets.filters.heuristic.is_prefixed_with_dollar_sign',
        },
        {
            'path': 'detect_secrets.filters.heuristic.is_indirect_reference',
        },
        {
            'path': 'detect_secrets.filters.common.is_ignored_due_to_verification_policies',

            # Hard-code this, just in case VerifiedResult enum values changes.
            # This corresponds to VerifiedResult.UNVERIFIED
            'min_level': 2,
        },
    ]

    if baseline.get('exclude'):
        if baseline['exclude'].get('files'):
            baseline['filters_used'].append({
                'path': 'detect_secrets.filters.regex.should_exclude_file',
                'pattern': [
                    baseline['exclude']['files'],
                ],
            })

        if baseline['exclude'].get('lines'):
            baseline['filters_used'].append({
                'path': 'detect_secrets.filters.regex.should_exclude_line',
                'pattern': [
                    baseline['exclude']['lines'],
                ],
            })

        baseline.pop('exclude')

    if baseline.get('word_list'):
        if baseline['word_list']['file']:
            baseline['filters_used'].append({
                'path': 'detect_secrets.filters.wordlist.should_exclude_secret',
                'min_length': 3,
                'file_name': baseline['word_list']['file'],
                'file_hash': baseline['word_list']['hash'],
            })

        baseline.pop('word_list')


def _rename_high_entropy_string_arguments(baseline: Dict[str, Any]) -> None:
    """
    During the great refactor for v1.0.0, we also decided to rename these arguments for
    consistency and simplicity.
    """
    for plugin in baseline['plugins_used']:
        if plugin['name'] == 'Base64HighEntropyString':
            plugin['limit'] = plugin.pop('base64_limit')

        elif plugin['name'] == 'HexHighEntropyString':
            plugin['limit'] = plugin.pop('hex_limit')

    # TODO: KeywordDetector?


def _migrate_custom_plugins(baseline: Dict[str, Any]) -> None:
    if 'custom_plugin_paths' not in baseline:
        return

    for path in baseline['custom_plugin_paths']:
        try:
            # NOTE: We don't want to use `detect_secrets.core.plugins.initialize.from_file`
            # since we don't want to *initialize* these plugins. That will pollute our global
            # settings object. Instead, we're merely "parsing" this file, and applying changes
            # as necessary.
            custom_plugins = cast(Iterable, get_plugins_from_file(path))
        except InvalidFile:
            # Best effort upgrade. Don't break if invalid file.
            continue

        for plugin in custom_plugins:
            baseline['plugins_used'].append({
                'name': plugin.__name__,
                'path': f'file://{os.path.abspath(path)}',
            })

    del baseline['custom_plugin_paths']



"""
This plugin searches for NPM tokens
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class NpmDetector(RegexBasedDetector):
    """Scans for NPM tokens."""
    secret_type = 'NPM tokens'

    denylist = [
        # npmrc authToken
        # ref. https://stackoverflow.com/questions/53099434/using-auth-tokens-in-npmrc
        re.compile(
            r'(?:npm_[A-Za-z0-9-]{36})|' +
            r'(?:(?:_authToken|NPM[ _-]?TOKEN)[\s\S]{0,5}?(?:NpmToken\.)?([a-z0-9-]{36}).*)',
        ),
    ]



"""
This code was extracted in part from
https://github.com/PyCQA/bandit. Using similar heuristic logic,
we adapted it to fit our plugin infrastructure, to create an organized,
concerted effort in detecting all type of secrets in code.

Copyright (c) 2014 Hewlett-Packard Development Company, L.P.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from __future__ import annotations

import re
from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Pattern
from typing import Set

from ..core.potential_secret import PotentialSecret
from ..util.filetype import determine_file_type
from ..util.filetype import FileType
from .base import BasePlugin
from detect_secrets.util.code_snippet import CodeSnippet


# Note: All values here should be lowercase
DENYLIST = (
    'api[_\\.\\-]?key',
    'auth_?key',
    'service_?key',
    'account_?key',
    'db_?key',
    'database_?key',
    'priv_?key',
    'private_?key',
    'client_?key',
    'db_?pass',
    'database_?pass',
    'key_?pass',
    'password',
    'passwd',
    'token',
    '_pass\\b',
    'pwd',
    'secret',
    'contraseña',
    'contrasena',
    'recaptcha_.*key',
    'nessus_?key',
)
# Includes ], ', " as closing
CLOSING = r'[]\'"]{0,2}'
AFFIX_REGEX = r'\w*'
DENYLIST_REGEX = r'|'.join(DENYLIST)
# Support for suffix after keyword i.e. password_secure = "value"
DENYLIST_REGEX = r'({denylist}){suffix}'.format(
    denylist=DENYLIST_REGEX,
    suffix=AFFIX_REGEX,
)
# Support for prefix and suffix with keyword, needed for reverse comparisons
# i.e. if ("value" == my_password_secure) {}
DENYLIST_REGEX_WITH_PREFIX = r'{prefix}{denylist}'.format(
    prefix=AFFIX_REGEX,
    denylist=DENYLIST_REGEX,
)
# Non-greedy match
OPTIONAL_WHITESPACE = r'\s*'
OPTIONAL_NON_WHITESPACE = r'[^\s]{0,50}?'
QUOTE = r'[\'"`]'
# Secret regex details:
#   (?=[^\v\'"]*)   ->  this section match with every character except line breaks and quotes. This
#                       allows to find secrets that starts with symbols or alphanumeric characters.
#
#   (?=\w+)     ->  this section match only with words (letters, numbers or _ are allowed), and at
#                   least one character is required. This allows to reduce the false positives
#                   number.
#
#   [^\v\'"]*   ->  this section match with every character except line breaks and quotes. This
#                   allows to find secrets with symbols at the end.
#
#   [^\v,\'"`]  ->  this section match with the last secret character that can be everything except
#                   line breaks, comma, backticks or quotes. This allows to reduce the false
#                   positives number and to prevent errors in the code snippet highlighting.

# '$' sign is being used for variable rendering. That why it's not here.
SECRET = r'(?=[^\v\'\"]*)(?=[a-zA-Z0-9_!@#%^&])[^\v\'\"]*[^\v,\'\"`]'
SQUARE_BRACKETS = r'(\[[0-9]*\])'

FOLLOWED_BY_COLON_EQUAL_SIGNS_REGEX = re.compile(
    # e.g. my_password := "bar" or my_password := bar
    r'{denylist}({closing})?{whitespace}:=?{whitespace}({quote}?)({secret})(\3)'.format(
        denylist=DENYLIST_REGEX,
        closing=CLOSING,
        quote=QUOTE,
        whitespace=OPTIONAL_WHITESPACE,
        secret=SECRET,
    ),
    flags=re.IGNORECASE,
)
FOLLOWED_BY_COLON_REGEX = re.compile(
    # e.g. api_key: foo
    r'{denylist}({closing})?:{whitespace}({quote}?)({secret})(\3)'.format(
        denylist=DENYLIST_REGEX,
        closing=CLOSING,
        quote=QUOTE,
        whitespace=OPTIONAL_WHITESPACE,
        secret=SECRET,
    ),
    flags=re.IGNORECASE,
)

FOLLOWED_BY_COLON_QUOTES_REQUIRED_REGEX = re.compile(
    # e.g. api_key: "foo"
    r'{denylist}({closing})?:({whitespace})({quote})({secret})(\4)'.format(
        denylist=DENYLIST_REGEX,
        closing=CLOSING,
        quote=QUOTE,
        whitespace=OPTIONAL_WHITESPACE,
        secret=SECRET,
    ),
    flags=re.IGNORECASE,
)
FOLLOWED_BY_EQUAL_SIGNS_OPTIONAL_BRACKETS_OPTIONAL_AT_SIGN_QUOTES_REQUIRED_REGEX = re.compile(
    # e.g. my_password = "bar"
    # e.g. my_password = @"bar"
    # e.g. my_password[] = "bar";
    # e.g. char my_password[25] = "bar";
    r'{denylist}({square_brackets})?{optional_whitespace}[!=]{{1,2}}{optional_whitespace}(@)?(")({secret})(\5)'.format(
        denylist=DENYLIST_REGEX,
        square_brackets=SQUARE_BRACKETS,
        optional_whitespace=OPTIONAL_WHITESPACE,
        secret=SECRET,
    ),
    flags=re.IGNORECASE,
)
FOLLOWED_BY_OPTIONAL_ASSIGN_QUOTES_REQUIRED_REGEX = re.compile(
    # e.g. std::string secret("bar");
    # e.g. secret.assign("bar",17);
    r'{denylist}(.assign)?\((")({secret})(\3)'.format(
        denylist=DENYLIST_REGEX,
        secret=SECRET,
    ),
)
FOLLOWED_BY_EQUAL_SIGNS_REGEX = re.compile(
    # e.g. my_password = bar
    # e.g. my_password == "bar" or my_password != "bar" or my_password === "bar"
    # or my_password !== "bar"
    # e.g. my_password == 'bar' or my_password != 'bar' or my_password === 'bar'
    # or my_password !== 'bar'
    r'{denylist}({closing})?{whitespace}(={{1,3}}|!==?){whitespace}({quote}?)({secret})(\4)'.format(
        denylist=DENYLIST_REGEX,
        closing=CLOSING,
        quote=QUOTE,
        whitespace=OPTIONAL_WHITESPACE,
        secret=SECRET,
    ),
    flags=re.IGNORECASE,
)
FOLLOWED_BY_EQUAL_SIGNS_QUOTES_REQUIRED_REGEX = re.compile(
    # e.g. my_password = "bar"
    # e.g. my_password == "bar" or my_password != "bar" or my_password === "bar"
    # or my_password !== "bar"
    # e.g. my_password == 'bar' or my_password != 'bar' or my_password === 'bar'
    # or my_password !== 'bar'
    r'{denylist}({closing})?{whitespace}(={{1,3}}|!==?){whitespace}({quote})({secret})(\4)'.format(
        denylist=DENYLIST_REGEX,
        closing=CLOSING,
        quote=QUOTE,
        whitespace=OPTIONAL_WHITESPACE,
        secret=SECRET,
    ),
    flags=re.IGNORECASE,
)
PRECEDED_BY_EQUAL_COMPARISON_SIGNS_QUOTES_REQUIRED_REGEX = re.compile(
    # e.g. "bar" == my_password or "bar" != my_password or "bar" === my_password
    # or "bar" !== my_password
    # e.g. 'bar' == my_password or 'bar' != my_password or 'bar' === my_password
    # or 'bar' !== my_password
    r'({quote})({secret})(\1){whitespace}[!=]{{2,3}}{whitespace}{denylist}'.format(
        denylist=DENYLIST_REGEX_WITH_PREFIX,
        quote=QUOTE,
        whitespace=OPTIONAL_WHITESPACE,
        secret=SECRET,
    ),
)
FOLLOWED_BY_QUOTES_AND_SEMICOLON_REGEX = re.compile(
    # e.g. private_key "something";
    r'{denylist}{nonWhitespace}{whitespace}({quote})({secret})(\2);'.format(
        denylist=DENYLIST_REGEX,
        nonWhitespace=OPTIONAL_NON_WHITESPACE,
        quote=QUOTE,
        whitespace=OPTIONAL_WHITESPACE,
        secret=SECRET,
    ),
    flags=re.IGNORECASE,
)
FOLLOWED_BY_ARROW_FUNCTION_SIGN_QUOTES_REQUIRED_REGEX = re.compile(
    # e.g. my_password => "bar" or my_password => bar
    r'{denylist}({closing})?{whitespace}=>?{whitespace}({quote})({secret})(\3)'.format(
        denylist=DENYLIST_REGEX,
        closing=CLOSING,
        quote=QUOTE,
        whitespace=OPTIONAL_WHITESPACE,
        secret=SECRET,
    ),
    flags=re.IGNORECASE,
)
CONFIG_DENYLIST_REGEX_TO_GROUP = {
    FOLLOWED_BY_COLON_REGEX: 4,
    PRECEDED_BY_EQUAL_COMPARISON_SIGNS_QUOTES_REQUIRED_REGEX: 2,
    FOLLOWED_BY_EQUAL_SIGNS_REGEX: 5,
    FOLLOWED_BY_QUOTES_AND_SEMICOLON_REGEX: 3,
}
GOLANG_DENYLIST_REGEX_TO_GROUP = {
    FOLLOWED_BY_COLON_EQUAL_SIGNS_REGEX: 4,
    PRECEDED_BY_EQUAL_COMPARISON_SIGNS_QUOTES_REQUIRED_REGEX: 2,
    FOLLOWED_BY_EQUAL_SIGNS_REGEX: 5,
    FOLLOWED_BY_QUOTES_AND_SEMICOLON_REGEX: 3,
}
COMMON_C_DENYLIST_REGEX_TO_GROUP = {
    FOLLOWED_BY_EQUAL_SIGNS_OPTIONAL_BRACKETS_OPTIONAL_AT_SIGN_QUOTES_REQUIRED_REGEX: 6,
}
C_PLUS_PLUS_REGEX_TO_GROUP = {
    FOLLOWED_BY_OPTIONAL_ASSIGN_QUOTES_REQUIRED_REGEX: 4,
    FOLLOWED_BY_EQUAL_SIGNS_QUOTES_REQUIRED_REGEX: 5,
}
QUOTES_REQUIRED_DENYLIST_REGEX_TO_GROUP = {
    FOLLOWED_BY_COLON_QUOTES_REQUIRED_REGEX: 5,
    PRECEDED_BY_EQUAL_COMPARISON_SIGNS_QUOTES_REQUIRED_REGEX: 2,
    FOLLOWED_BY_EQUAL_SIGNS_QUOTES_REQUIRED_REGEX: 5,
    FOLLOWED_BY_QUOTES_AND_SEMICOLON_REGEX: 3,
    FOLLOWED_BY_ARROW_FUNCTION_SIGN_QUOTES_REQUIRED_REGEX: 4,
}

TERRAFORM_DENYLIST_REGEX_TO_GROUP = {
    **QUOTES_REQUIRED_DENYLIST_REGEX_TO_GROUP,
    FOLLOWED_BY_EQUAL_SIGNS_REGEX: 5,
}

REGEX_BY_FILETYPE = {
    FileType.GO: GOLANG_DENYLIST_REGEX_TO_GROUP,
    FileType.OBJECTIVE_C: COMMON_C_DENYLIST_REGEX_TO_GROUP,
    FileType.C_SHARP: COMMON_C_DENYLIST_REGEX_TO_GROUP,
    FileType.C: COMMON_C_DENYLIST_REGEX_TO_GROUP,
    FileType.C_PLUS_PLUS: C_PLUS_PLUS_REGEX_TO_GROUP,
    FileType.CLS: QUOTES_REQUIRED_DENYLIST_REGEX_TO_GROUP,
    FileType.JAVA: QUOTES_REQUIRED_DENYLIST_REGEX_TO_GROUP,
    FileType.JAVASCRIPT: QUOTES_REQUIRED_DENYLIST_REGEX_TO_GROUP,
    FileType.PYTHON: QUOTES_REQUIRED_DENYLIST_REGEX_TO_GROUP,
    FileType.SWIFT: QUOTES_REQUIRED_DENYLIST_REGEX_TO_GROUP,
    FileType.TERRAFORM: TERRAFORM_DENYLIST_REGEX_TO_GROUP,
    FileType.JSON: QUOTES_REQUIRED_DENYLIST_REGEX_TO_GROUP,
    FileType.YAML: CONFIG_DENYLIST_REGEX_TO_GROUP,
    FileType.CONFIG: CONFIG_DENYLIST_REGEX_TO_GROUP,
    FileType.INI: CONFIG_DENYLIST_REGEX_TO_GROUP,
    FileType.PROPERTIES: CONFIG_DENYLIST_REGEX_TO_GROUP,
    FileType.TOML: CONFIG_DENYLIST_REGEX_TO_GROUP,
}


class KeywordDetector(BasePlugin):
    """
    Scans for secret-sounding variable names.

    This checks if denylisted keywords are present in the analyzed string.
    """
    secret_type = 'Secret Keyword'

    def __init__(self, keyword_exclude: Optional[str] = None) -> None:
        self.keyword_exclude = None
        if keyword_exclude:
            self.keyword_exclude = re.compile(
                keyword_exclude,
                re.IGNORECASE,
            )

    def analyze_string(
        self,
        string: str,
        denylist_regex_to_group: Optional[Dict[Pattern, int]] = None,
    ) -> Generator[str, None, None]:
        if self.keyword_exclude and self.keyword_exclude.search(string):
            return

        if denylist_regex_to_group is None:
            attempts = [
                QUOTES_REQUIRED_DENYLIST_REGEX_TO_GROUP,
            ]
        else:
            attempts = [denylist_regex_to_group]

        has_results = False
        for denylist_regex_to_group in attempts:
            for denylist_regex, group_number in denylist_regex_to_group.items():
                match = denylist_regex.search(string)
                if match:
                    has_results = True
                    yield match.group(group_number)

            if has_results:
                break

    def analyze_line(
        self,
        filename: str,
        line: str,
        line_number: int = 0,
        context: CodeSnippet | None = None,
        raw_context: CodeSnippet | None = None,
        **kwargs: Any,
    ) -> Set[PotentialSecret]:
        filetype = determine_file_type(filename)
        denylist_regex_to_group = REGEX_BY_FILETYPE.get(filetype, QUOTES_REQUIRED_DENYLIST_REGEX_TO_GROUP)  # noqa: E501
        return super().analyze_line(
            filename=filename,
            line=line,
            line_number=line_number,
            context=context,
            raw_context=raw_context,
            denylist_regex_to_group=denylist_regex_to_group,
        )

    def json(self) -> Dict[str, Any]:
        return {
            'keyword_exclude': (
                self.keyword_exclude.pattern
                if self.keyword_exclude
                else ''
            ),
            **super().json(),
        }



"""
There may be known words that are definitely test keys (e.g. AKIATEST for AWS keys).
One way that we can filter these out is by passing in a list of words that we know
will result in false positives. This filter efficiently processes this through the
use of the Aho-Corasick algorithm.
"""
from functools import lru_cache
from typing import Any

from ..settings import get_settings
from .util import compute_file_hash


Automaton = Any


def is_feature_enabled() -> bool:
    try:
        get_automaton()
        return True
    except ImportError:
        return False


def initialize(wordlist_filename: str, min_length: int = 3, file_hash: str = '') -> Automaton:  #noqa: ARG001
    """
    :param min_length: if words are too small, the automaton will flag too many
        words. As a result, our recall will decrease without a precision boost.
        Tweak this value to customize it based on your own findings.

    :param file_hash: this is currently used for baseline reporting purposes only, rather than
        engine's functionality. One can imagine a future where this automaton model is
        cached and keyed off the hash, and thus, this file_hash can be used to see if the
        cache needs to be invalidated.

        But alas, this functionality has yet to be implemented.
    """
    # See https://pyahocorasick.readthedocs.io/en/latest/ for more information.
    automaton = get_automaton()
    with open(wordlist_filename) as f:
        for line in f.readlines():
            line = line.lower().strip()

            if len(line) < min_length:
                continue

            automaton.add_word(line, line)

    path = f'{__name__}.should_exclude_secret'
    get_settings().filters[path] = {
        'min_length': min_length,
        'file_name': wordlist_filename,
        'file_hash': compute_file_hash(wordlist_filename),
    }

    automaton.make_automaton()
    return automaton


def should_exclude_secret(secret: str) -> bool:
    try:
        # .lower() to make everything case-insensitive
        next(get_automaton().iter(string=secret.lower()))
        return True
    except StopIteration:
        return False


@lru_cache(maxsize=1)
def get_automaton() -> Automaton:
    import ahocorasick  # type:ignore[import-not-found]
    return ahocorasick.Automaton()



from . import analytics                 # noqa: F401
from . import report                    # noqa: F401
from .audit import audit_baseline       # noqa: F401
from .compare import compare_baselines  # noqa: F401



"""
This plugin searches for Azure Storage Account access keys.
"""
from __future__ import annotations

import re
from typing import Any
from typing import List
from typing import Optional
from typing import Set

from detect_secrets.core.potential_secret import PotentialSecret
from detect_secrets.plugins.base import RegexBasedDetector
from detect_secrets.util.code_snippet import CodeSnippet


class AzureStorageKeyDetector(RegexBasedDetector):
    """Scans for Azure Storage Account access keys."""
    secret_type = 'Azure Storage Account access key'

    account_key = 'AccountKey'
    azure = 'azure'

    max_line_length = 4000
    max_part_length = 2000
    integrity_regex = re.compile(r'integrity[:=]')

    denylist = [
        # Account Key (AccountKey=xxxxxxxxx)
        re.compile(
            r'(?:["\']?[A-Za-z0-9+\/]{86,1000}==["\']?)',
        ),
    ]

    context_keys = [
        r'{account_key}=\s*{secret}',

        # maximum 2 lines secret distance under azure mention (case-insensitive)
        r'(?i)\b{azure}(.*\n){{0,2}}.*{secret}',

        # maximum 2 lines secret distance above azure mention (case-insensitive)
        r'(?i)\b{secret}(.*\n){{0,2}}.*{azure}',
    ]

    def analyze_line(
            self,
            filename: str,
            line: str,
            line_number: int = 0,
            context: Optional[CodeSnippet] = None,
            raw_context: Optional[CodeSnippet] = None,
            **kwargs: Any,
    ) -> Set[PotentialSecret]:
        output: Set[PotentialSecret] = set()
        results = super().analyze_line(
            filename=filename, line=line, line_number=line_number,
            context=context, raw_context=raw_context, **kwargs,
        )
        output.update(self.analyze_context_keys(results, context, line))

        return output

    def analyze_context_keys(
            self,
            results: Set[PotentialSecret],
            context: Optional[CodeSnippet],
            line: str,
    ) -> List[PotentialSecret]:
        context_text = '\n'.join(context.lines).replace('\n\n', '\n') if context else line
        return [result for result in results if self.context_keys_exists(result, context_text)]

    def context_keys_exists(self, result: PotentialSecret, string: str) -> bool:
        if len(string) > self.max_line_length:
            # for very long lines, we don't run the regex to avoid performance issues
            return False
        if result.secret_value:
            for secret_regex in self.context_keys:
                regex = re.compile(
                    secret_regex.format(
                        secret=re.escape(result.secret_value), account_key=self.account_key,
                        azure=self.azure,
                    ), re.MULTILINE,
                )
                if regex.pattern.startswith(self.account_key) and self.account_key not in string:
                    continue
                if self.azure in regex.pattern.lower() and self.azure not in string.lower():
                    continue
                if self.contains_integrity(result.secret_value, string):
                    continue
                if regex.search(string) is not None:
                    return True
        return False

    def contains_integrity(self, secret_val: str, string: str) -> bool:
        # we want to ignore cases of lock files which contains hashes
        context_parts = string.split('\n')
        return any(
            len(part) < self.max_part_length and
            secret_val in part and
            self.integrity_regex.search(part) is not None for part in context_parts
        )



"""
The compare module allows developers/analysts to determine the difference between two
different baseline files. This allows easier configuration of plugin settings.

For effective use, a few assumptions are made:
    1. Baselines are sorted by (filename, line_number, hash).
        This allows for a deterministic order, when doing a side-by-side
        comparison.

    2. Baselines are generated for the same codebase snapshot.
        This means that we won't have cases where secrets are moved around;
        only added or removed.

NOTE: We don't want to do a version check, because we want to be able to
use this functionality across versions (to see how the new version fares
compared to the old one).
"""
from typing import Any
from typing import Dict
from typing import Iterator
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

from . import io
from ..core import baseline
from ..core.potential_secret import PotentialSecret
from ..core.secrets_collection import SecretsCollection
from ..custom_types import SecretContext
from ..exceptions import NoLineNumberError
from ..exceptions import SecretNotFoundOnSpecifiedLineError
from ..settings import transient_settings
from ..util.code_snippet import get_code_snippet
from ..util.color import AnsiColor
from ..util.color import colorize
from .common import get_raw_secret_from_file
from .common import open_file
from .iterator import BidirectionalIterator


def compare_baselines(old_baseline_filename: str, new_baseline_filename: str) -> None:
    if old_baseline_filename == new_baseline_filename:
        io.print_error('This is the same file!')
        return

    old_baseline, old_config = _get_baseline_from_file(old_baseline_filename)
    new_baseline, new_config = _get_baseline_from_file(new_baseline_filename)

    # We want to display all entries, so make sure all files are present.
    old_baseline.trim()
    new_baseline.trim()

    try:
        _display_difference_to_user((old_baseline, old_config), (new_baseline, new_config))
    except NoLineNumberError as e:
        io.print_error(str(e))


def _get_baseline_from_file(filename: str) -> Tuple[SecretsCollection, Dict[str, Any]]:
    data = baseline.upgrade(baseline.load_from_file(filename))
    config = {
        'plugins_used': data['plugins_used'],
        'filters_used': [] if 'filters_used' not in data else data['filters_used'],
    }

    return baseline.load(data, filename), config


def _compare_baselines(
    old_baseline: SecretsCollection,
    new_baseline: SecretsCollection,
) -> Iterator[Tuple[str, Optional[PotentialSecret], Optional[PotentialSecret]]]:
    """
    :returns: (filename, left_secret, right_secret)
        `filename` is needed to know which file to display;
        if `left_secret` is None, then it's a newly added secret;
        if `right_secret` is None, then it's a deleted secret
    """
    class LeftSecret(Exception):  # noqa: N818
        pass

    class RightSecret(Exception):  # noqa: N818
        pass

    left_secrets = [secret for _, secret in old_baseline]
    right_secrets = [secret for _, secret in new_baseline]

    left_index = 0
    right_index = 0
    while left_index < len(left_secrets) or right_index < len(right_secrets):
        try:
            # This allows us to delay execution of the exception handling, until we had a chance
            # to initialize both variables. Either one must at least pass, otherwise the while
            # statement will be False.
            exception: Optional[Union[Type[LeftSecret], Type[RightSecret]]] = None
            try:
                left_secret = left_secrets[left_index]
                if not left_secret.line_number:
                    raise NoLineNumberError
            except IndexError:
                exception = RightSecret

            try:
                right_secret = right_secrets[right_index]
                if not right_secret.line_number:
                    raise NoLineNumberError
            except IndexError:
                exception = LeftSecret

            if exception:
                raise exception

            # At this point, both secrets exist. So now, we'll check for filenames.
            # If they aren't talking about the same file, then the rest of the secrets in the
            # file must not exist.
            if left_secret.filename < right_secret.filename:
                raise LeftSecret
            elif left_secret.filename > right_secret.filename:
                raise RightSecret

            # When it's the same file, we want to show it by line order of the combined
            # list.
            if left_secret.line_number < right_secret.line_number:
                raise LeftSecret
            elif left_secret.line_number > right_secret.line_number:
                raise RightSecret

            # At this point, it's the same line number, and the same filename.
            # They could be different secrets (and therefore, different places on the same line)
            # but we'll just show the alphabetically lowest one first (easier implementation).
            if left_secret.secret_hash < right_secret.secret_hash:
                raise LeftSecret
            elif left_secret.secret_hash > right_secret.secret_hash:
                raise RightSecret

            # At this point, they must be referencing the same secret value (not necessarily
            # same *secret*, since it could be detected by different plugins). However, since this
            # is a comparison, we're going to currently ignore if the secret values are the same.
            old_hash = left_secret.secret_hash
            try:
                while old_hash == left_secrets[left_index].secret_hash:
                    left_index += 1
            except IndexError:
                pass

            old_hash = right_secret.secret_hash
            try:
                while old_hash == right_secrets[right_index].secret_hash:
                    right_index += 1
            except IndexError:
                pass

        except LeftSecret:
            yield (left_secret.filename, left_secret, None)
            left_index += 1

        except RightSecret:
            yield (right_secret.filename, None, right_secret)
            right_index += 1


def _display_difference_to_user(
    old_data: Tuple[SecretsCollection, Dict[str, Any]],
    new_data: Tuple[SecretsCollection, Dict[str, Any]],
) -> None:
    """
    :param old_data: it needs both, since it can technically be scanning two different
        plugins' settings.
    :param new_data: same as `old_data`
    """
    old_baseline, old_config = old_data
    new_baseline, new_config = new_data

    iterator = BidirectionalIterator(list(_compare_baselines(old_baseline, new_baseline)))
    for _, left_secret, right_secret in iterator:
        io.clear_screen()

        secret = left_secret if left_secret else right_secret
        config = old_config if left_secret else new_config

        try:
            with transient_settings(config):
                secret.secret_value = get_raw_secret_from_file(secret)

            context = SecretContext(
                current_index=iterator.index + 1,
                num_total_secrets=len(iterator.collection),
                secret=secret,
                header='{status}      {value}'.format(
                    status=colorize('Status:', AnsiColor.BOLD),
                    value='>> {} <<'.format(
                        colorize('REMOVED', AnsiColor.RED)
                        if not right_secret
                        else colorize('ADDED', AnsiColor.LIGHT_GREEN),
                    ),
                ),
                snippet=get_code_snippet(
                    lines=open_file(secret.filename).raw_lines,
                    line_number=secret.line_number,
                ),
            )
            io.print_context(context)

            decision = io.get_user_decision(
                can_step_back=iterator.can_step_back(),
                prompt_secret_decision=False,
            )
        except SecretNotFoundOnSpecifiedLineError as e:
            io.print_secret_not_found(
                SecretContext(
                    current_index=iterator.index + 1,
                    num_total_secrets=len(iterator.collection),
                    secret=secret,
                    error=e,
                ),
            )

            decision = io.get_user_decision(
                prompt_secret_decision=False,
                can_step_back=iterator.can_step_back(),
            )

        if decision == io.InputOptions.QUIT:
            io.print_message('Quitting...')
            break
        elif decision == io.InputOptions.BACK:
            iterator.step_back_on_next_iteration()



import json
import time
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from . import upgrades
from ..__version__ import VERSION
from ..exceptions import UnableToReadBaselineError
from ..settings import configure_settings_from_baseline
from ..settings import get_settings
from ..util.importlib import import_modules_from_package
from ..util.semver import Version
from .scan import get_files_to_scan
from .secrets_collection import SecretsCollection


def create(
    *paths: str,
    should_scan_all_files: bool = False,
    root: str = '',
    num_processors: Optional[int] = None,
) -> SecretsCollection:
    """Scans all the files recursively in path to initialize a baseline."""
    kwargs = {}
    if num_processors:
        kwargs['num_processors'] = num_processors

    secrets = SecretsCollection(root=root)
    secrets.scan_files(
        *get_files_to_scan(*paths, should_scan_all_files=should_scan_all_files, root=root),
        **kwargs,
    )

    return secrets


def load(baseline: Dict[str, Any], filename: str = '') -> SecretsCollection:
    """
    With a given baseline file, load all settings and discovered secrets from it.

    :raises: KeyError
    """
    # This is required for backwards compatibility, and supporting upgrades from older versions.
    baseline = upgrade(baseline)

    configure_settings_from_baseline(baseline, filename=filename)
    return SecretsCollection.load_from_baseline(baseline)


def load_from_file(filename: str) -> Dict[str, Any]:
    """
    :raises: UnableToReadBaselineError
    :raises: InvalidBaselineError
    """
    try:
        with open(filename) as f:
            return cast(Dict[str, Any], json.loads(f.read()))
    except (FileNotFoundError, OSError, json.decoder.JSONDecodeError) as e:
        raise UnableToReadBaselineError from e


def format_for_output(secrets: SecretsCollection, is_slim_mode: bool = False) -> Dict[str, Any]:
    output = {
        'version': VERSION,

        # This will populate settings of filters and plugins,
        **get_settings().json(),

        'results': secrets.json(),
    }

    if not is_slim_mode:
        output['generated_at'] = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())
    else:
        # NOTE: This has a nice little side effect of keeping it ordered by line number,
        # even though we don't output it.
        for _, secret_list in cast(
            Dict[str, List[Dict[str, Any]]],
            output['results'],
        ).items():
            for secret_dict in secret_list:
                secret_dict.pop('line_number')

    return output


def save_to_file(
    secrets: Union[SecretsCollection, Dict[str, Any]],
    filename: str,
) -> None:    # pragma: no cover
    """
    :param secrets: if this is a SecretsCollection, it will output the baseline in its latest
        format. Otherwise, you should pass in a dictionary to this function, to manually
        specify the baseline format to save as.

        If you're trying to decide the difference, ask yourself whether there are any changes
        that does not directly impact the results of the scan.
    """
    # TODO: I wonder whether this should add the `detect_secrets.filters.common.is_baseline_file`
    # filter, since we know the filename already. However, one could argue that it would cause
    # this function to "do more than one thing".
    output = secrets
    if isinstance(secrets, SecretsCollection):
        output = format_for_output(secrets)

    with open(filename, 'w') as f:
        f.write(json.dumps(output, indent=2) + '\n')


def upgrade(baseline: Dict[str, Any]) -> Dict[str, Any]:
    """
    Baselines will eventually require format changes. This function is responsible for upgrading
    an older baseline to the latest version.
    """
    baseline_version = Version(baseline['version'])
    if baseline_version >= Version(VERSION):
        return baseline

    modules = import_modules_from_package(
        upgrades,
        filter=lambda x: not _is_relevant_upgrade_module(baseline_version)(x),
    )

    new_baseline = {**baseline}
    for module in modules:
        module.upgrade(new_baseline)

    new_baseline['version'] = VERSION
    return new_baseline


def _is_relevant_upgrade_module(current_version: Version) -> Callable:
    def wrapped(module_path: str) -> bool:
        # This converts `v1_0` to `1.0`
        affected_version_string = module_path.rsplit('.', 1)[-1].lstrip('v').replace('_', '.')

        # Patch version doesn't matter, because patches should not require baseline bumps.
        affected_version = Version(f'{affected_version_string}.0')

        return current_version < affected_version

    return wrapped
